export type Chunk = {
  highlight: boolean,
  start: number,
  end: number,
  className?: string,
  key?: number,
  value?: string
};



/**
 * 
 * @param string 
 */
function defaultSanitize (string: string): string {
  return string;
}

/**
 * 处理特殊字符
 * @param string 
 */
function escapeRegExpFn (string: string): string {
  // eslint-disable-next-line no-useless-escape
  return string.replace(/[\-\[\]\*\&\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&');
}

/**
 * 对所有切割的模块进行同类交叉、异类交叉、异类包含处理
 * @return {Chunk[]}
 */
export const combineChunks = ({
  chunks,
}: {
  chunks: Array<Chunk>,
}): Array<Chunk> => {
  chunks = chunks
    .sort((first: Chunk, second: Chunk) => first.start - second.start)
    .reduce((processedChunks: Array<Chunk>, nextChunk: Chunk) => {
      // 第一个模块
      if (processedChunks.length === 0) {
        return [nextChunk];
      } 


      // 已处理的最后项
      const prevChunk: Chunk | undefined = processedChunks.pop();
      if (typeof prevChunk === 'undefined') {
        throw new Error('processedChunks is empty')
      }
      // 可能会有没有交集的情况
      if (nextChunk.end <= prevChunk.start) {
        processedChunks.push({ ...prevChunk });
        return processedChunks;
      }

      /**
       * 如果相邻的两段存在交叉内容这合并，
       * 当只有一种高亮时，直接合并即可
       * 当存在两种类型时，在有交叉的情况下：
       * 1、如果是同类型，那么直接合并
       * 2、如果是不同的类型，获取的部分，相邻的两段只保留自己独立的部分
       */
      if (nextChunk.start <= prevChunk.end) {
        // 相交叉或者完全重复 默认文本长的在前
        if (nextChunk.className !== prevChunk.className) {


          const start = prevChunk.start;
          const end = Math.max(prevChunk.end, nextChunk.end);
          const commonStart = Math.max(prevChunk.start, nextChunk.start);
          const commonEnd = Math.min(prevChunk.end, nextChunk.end);



          if (nextChunk.start === prevChunk.start && nextChunk.end === prevChunk.end) {
            processedChunks.push({
              ...nextChunk,
              className: `${nextChunk.className} ${prevChunk.className}`
            });
          } else if (nextChunk.start <= prevChunk.start) {
            processedChunks.push(
              {
                className: `${nextChunk.className} ${prevChunk.className}`,
                start,
                end: commonEnd,
                highlight: false
              },
              {
                ...prevChunk,
                start: commonEnd,
                end,
                // 内容较长的独立部分className
                className: nextChunk.end > prevChunk.end ? nextChunk.className : prevChunk.className
              }
            );
          } else if (nextChunk.end === prevChunk.end) {
            processedChunks.push(
              {
                ...prevChunk,
                start,
                end: commonStart
              },
              {
                ...nextChunk,
                start: commonStart,
                end: commonEnd,
                className: `${nextChunk.className} ${prevChunk.className}`
              }
            );
          } else {
            processedChunks.push(
              {
                ...prevChunk,
                start,
                end: commonStart,
                className: nextChunk.start < prevChunk.start ? nextChunk.className : prevChunk.className
              },
              {
                highlight: false, 
                start: commonStart, 
                end: commonEnd,
                className: `${nextChunk.className} ${prevChunk.className}`
              },
              {
                highlight: false, 
                start: commonEnd, 
                end,
                className: nextChunk.end > prevChunk.end ? nextChunk.className : prevChunk.className
              }
            );
          }

        } else {
          // 如果是同类型匹配，则合并处理，end为两者end最大值
          const end = Math.max(prevChunk.end, nextChunk.end);
          processedChunks.push({
            ...nextChunk,
            end,
            highlight: false, 
            start: prevChunk.start 
          });
        }
      } else {
        // 完全独立的两项
        processedChunks.push(prevChunk, nextChunk);
      }

      return processedChunks;
      
    }, []);
  return chunks;
};

/**
 * 截取高亮文本，并获取文本位置
 * @param param0 
 */
const execWordPosition = ({
  autoEscape,
  caseSensitive,
  sanitize = defaultSanitize,
  searchWords,
  textToHighlight,
  className
}:{
  autoEscape?: boolean,
  caseSensitive?: boolean,
  sanitize?: typeof defaultSanitize,
  searchWords: Array<string>,
  textToHighlight: string,
  className?: string
}): Array<Chunk> => {
  textToHighlight = sanitize(textToHighlight);

  /**
	 * 对内容进行去重，过滤无效内容
	 */
  return [...new Set([...searchWords])]
    .filter(searchWord => searchWord) // 移除空的字符串
    .reduce((chunks: Array<Chunk>, searchWord) => {
      searchWord = sanitize(searchWord);

      if (autoEscape) {
        // 默认处理特殊字符
        searchWord = escapeRegExpFn(searchWord); 
      }

      // 利用exec的全局匹配
      const regex = new RegExp(searchWord, caseSensitive ? 'g' : 'gi');
      let match;
      while ((match = regex.exec(textToHighlight))) {
        const start = match.index;
        const end = regex.lastIndex;
        if (end > start) {
          chunks.push({ highlight: false, start, end, className, value: match[0] });
        }

        // http://www.regexguru.com/2008/04/watch-out-for-zero-length-matches/
        if (match.index === regex.lastIndex) {
          regex.lastIndex++;
        }
      }
      return chunks;
    }, []);
};

/**
 * 
 * 处理words的中多组数据并合并在一起，然后交个combineChunks, combineChunks会利用其下标重新排序
 * @return {start:number, end:number, className: string}[]
 */
const defaultFindChunks = ({
  words,
  activeClasses = [],
  ...rest
}: {
  autoEscape?: boolean,
  caseSensitive?: boolean,
  sanitize?: typeof defaultSanitize,
  textToHighlight: string,
  words: Array<string | Array<string>>,
  activeClasses: Array<string>
}): Array<Chunk> => {

  if (typeof words[0] === 'string') {
    const searchWords: any = words;
    return execWordPosition({
      ...rest,
      searchWords: searchWords,
      className: activeClasses[0] || 'first'
    });
  }
 
  return words.reduce((oneChunks: Array<Chunk>, searchWords: any, index: number) => {
    return oneChunks.concat(
      execWordPosition({
        ...rest,
        searchWords,
        className: activeClasses[index] || `style_${index}`
      })
    );
  }, []);

};



export { defaultFindChunks as findChunks };

/**
 * 将原文本拆成包含需要高亮和不需要高亮的全部模块
 * @param chunksToHighlight 所有处理之后需要高亮的模块
 * @param totalLength number
 * @return {start:number, end:number, highlight:boolean, className: string}[]
 */
export const fillInChunks = ({
  chunksToHighlight,
  totalLength
}: {
  chunksToHighlight: Array<Chunk>,
  totalLength: number,
}): Array<Chunk> => {
  const allChunks: Array<Chunk> = [];
  const append = (start:number, end:number, highlight: boolean, className?: string) => {
    if (end - start > 0) {
      allChunks.push({
        start,
        end,
        highlight,
        className,
        key: start
      });
    }
  };

  if (chunksToHighlight.length === 0) {
    append(0, totalLength, false);
  } else {
    let lastIndex = 0;
    chunksToHighlight.forEach((chunk) => {
      append(lastIndex, chunk.start, false); // 无需高亮的项
      append(chunk.start, chunk.end, true, chunk.className);
      lastIndex = chunk.end;
    });
    append(lastIndex, totalLength, false);
  }
  return allChunks;
};




export interface FindAllChunksProps {
  autoEscape?: boolean,
  caseSensitive?: boolean,
  findChunks?: typeof defaultFindChunks,
  sanitize?: typeof defaultSanitize,
  words: Array<Array<string> | string>
  activeClasses: Array<string>
  textToHighlight: string,
}

/**
 * @return 返回所有的chunks,包括高亮和不需要高亮全部 { start:number, end:number, highlight:boolean, className: string })
 */
export const findAll = ({
  autoEscape,
  caseSensitive = false,
  findChunks = defaultFindChunks,
  sanitize,
  textToHighlight,
  words,
  activeClasses,
}: FindAllChunksProps): Array<Chunk> => (
  fillInChunks({
    chunksToHighlight: combineChunks({
      chunks: findChunks({
        autoEscape,
        caseSensitive,
        sanitize,
        textToHighlight,
        words,
        activeClasses
      })
    }),
    totalLength: textToHighlight ? textToHighlight.length : 0
  })
);

